SystemOrganization addCategory: #Squeak2JS!

OrderedCollection subclass: #S2JNode
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JNode commentStamp: 'ag 11/18/2010 13:09' prior: 0!
This is only a named collection, that is supposed to hold further collections or strings to form a tree!

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:08'!
named: aString with: anOrderedCollection
	
	^ self new
		name: aString;
		addAll: anOrderedCollection;
		yourself! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:58'!
named: aString withoutNils: anOrderedCollection
	
	anOrderedCollection ifNil: [ ^ nil ].
	^ self
		named: aString
		with: (anOrderedCollection reject: [ :each | each isNil])! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/21/2010 19:15'!
named: aString withoutNilsNotEmpty: anOrderedCollection
	"Returns nil, if resulting node would be empty"
	
	| result |
	result := self named: aString withoutNils: anOrderedCollection.
	^ (result isNil or: [result isEmpty])
		ifTrue: [ nil ]
		ifFalse: [ result ]! !

!S2JNode methodsFor: 'comparing' stamp: 'ag 11/21/2010 19:19'!
= anObject
	
	^ super = anObject and: 
		[ name = anObject name ]! !

!S2JNode methodsFor: 'as yet unclassified' stamp: 'LT 12/1/2010 14:06:25.154'!
isArrayNode

	^ name = 'array' or: [name = 'byteArray' or: [name = 'braceExpression']]! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name
	
	^ name! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name: aString
	
	name := aString! !

!S2JNode methodsFor: 'printing' stamp: 'ag 11/19/2010 00:39'!
printString
	
	| result first |
	first := true.
	result := name, ' ('.
	self do: [ :each |
		first ifTrue: [first := false] ifFalse: [result := result, ', '].
		result := result,  each printString].
	^ result, ')'! !

!S2JNode methodsFor: 'as yet unclassified' stamp: 'LT 12/1/2010 13:57:31.81'!
separatedBy: aString

	self 
		inject: '' 
		into: 
			[ :sum :each | 
			sum isEmpty 
				ifTrue: [each asString] 
				ifFalse: [sum, aString, each asString]].! !

TestCase subclass: #S2JParserTest
	instanceVariableNames: 'currentRule'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParserTest class methodsFor: 'testing' stamp: 'ag 11/23/2010 21:38'!
findUnmatchableMethod
	"Search all methods in this image, print errors to the Transcript, notify every 500 methods"
	
	| num |
	num := 0.
	Smalltalk allClasses do: [ :eachClass |
		eachClass methodDict do: [ :eachMethod |
			[ S2JParser matchMethod: eachMethod.
			num := num + 1.
			(num \\ 500 = 0) ifTrue: [ Transcript cr; show: 'Matched ', num asString, ' methods.' ]]
				on: Error do: [ :err | Transcript cr; show: num asString, ': ', eachMethod name, ' (', err asString, ')' ]]].
	Transcript cr; show: 'Finished!! ', num, ' method have been matched!!'.! !

!S2JParserTest class methodsFor: 'testing' stamp: 'ag 11/30/2010 16:03'!
matchRandomMethods: aNumber
	"Match some random methods from the image. Return a String summarizing the result.
	Failed methods are printed to the Transcript."
	
	| runs successes |
	runs := successes := 0.
	aNumber timesRepeat: 
		[ | method |
			method := self randomMethod.
			method ifNotNil: [
				[ runs := runs + 1. S2JParser matchMethod: method. successes := successes + 1 ]
					on: Error do: [ :err | Transcript cr; cr; show: 'Failed to parse: ', method name, '. ', err ]]].
	^ runs asString, ' runs, ', successes asString, ' successes'! !

!S2JParserTest class methodsFor: 'helpers' stamp: 'ag 11/30/2010 16:07'!
randomMethod
	"Methods, that don't work, e.g. because they are too long, are excluded."
	
	| class result |
	result := nil.
	class := Smalltalk allClasses atRandom.
	class methodDict ifNotEmpty: [
		result := class methodDict atRandom.
		result = ((TheWorldMainDockingBar methodDict at: #terseGuideToSqueak) getSource)
			ifTrue: [ result := nil ]].
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/23/2010 12:21'!
assert: aString matches: aRule
	"In most cases, throws an error when failing (hard fail)"
	
	| result |
	result := S2JParser matchString: aString with: aRule.
	self assert: result notNil.
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:16'!
assert: aString matches: aRule andProduces: anObject
	"In most cases, throws an error when failing (hard fail). #= Checks collections deeply."
	
	| result |
	result := self assert: aString matches: aRule.
	self assert: result = anObject.
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 19:24'!
assert: aString matches: aRule andProducesNode: aCollection
	"Matches the string to the rule and compares the result to an S2JNode containing the given collection and named like the rule.
	(Follows a convention)."
	
	| result |
	result := self assert: aString matches: aRule.
	self assert: result = (S2JNode named: aRule asString with: aCollection).
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:18'!
assertFails: aString
	"Checks against the currentRule"
	
	self
		assertFails: aString
		toMatch: self currentRule.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:23'!
assertFails: aString toMatch: aRule
	
	self
		should: [ self assert: aString matches: aRule ]
		raise: OM2Fail
		description: 'Matching should have failed but didn''t!!'.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/30/2010 10:57'!
assertMatches: aString
	"Check against the currentRule"
	
	self assert: aString matches: self currentRule.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 20:06'!
assertMatches: aString andProduces: anObject
	"Check against the currentRule"
	
	^ self
		assert: aString
		matches: self currentRule
		andProduces: anObject! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 20:06'!
assertMatches: aString andProducesNode: aCollection
	"Check against the currentRule"
	
	^ self
		assert: aString
		matches: self currentRule
		andProducesNode: aCollection! !

!S2JParserTest methodsFor: 'accessing' stamp: 'ag 11/21/2010 14:14'!
currentRule
	^ currentRule! !

!S2JParserTest methodsFor: 'accessing' stamp: 'ag 11/21/2010 14:14'!
currentRule: aSymbol
	currentRule := aSymbol! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 20:37'!
node: aString
	
	^ S2JNode named: aString with: {}! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 19:27'!
node: aString with: aCollection
	
	^ S2JNode named: aString with: aCollection! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 12/1/2010 10:21'!
testArray
	
	self
		currentRule: #array;
		
		assertMatches: '()' andProduces: #();
		assertMatches: '(  )  ' andProduces: #();
		assertMatches: '(  1 2   3) ' andProduces: #(1 2 3);
		assertMatches: '(''aa'' 1 ''b1'' 32)' andProduces: {self node: 'string' with: { 'aa' }. 1. self node: 'string' with: { 'b1' }. 32};
		assertMatches: '(#abc     11    ''a''   ) ' andProduces: {#abc. 11. self node: 'string' with: { 'a' }};
		assertMatches: '(  ( 1 2 3    ''2'')   #a  ''c''   )' andProduces: { {1. 2. 3. self node: 'string' with: { '2' } }. #a. self node: 'string' with: { 'c' } };
		assertMatches: '($a     11    ''a''   #a2   #(1 2   $2) )  ' andProduces: { $a. 11. self node: 'string' with: { 'a' }. #a2. {1. 2. $2} };
		assertMatches: '( a )' andProduces: #( a );
		assertMatches: '( a b )' andProduces: #( a b );
		assertMatches: '(d ((b ((a))) (c)))' andProduces: #(d ((b ((a))) (c)));
		assertMatches: '( self a:bc == |? ?!!!! )  ' andProduces: #( self a:bc == |? ?!!!! );
		assertMatches: '( self #( aa 1 ) ( ''a'' #1 #ab: ab:))  ' andProduces: { #self .{ #aa. 1 }. { self node: 'string' with: { 'a' }. 1. #ab:. #ab: } } ;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '(';
		assertFails: '#()';
		assertFails: '#(';
		assertFails: '{1}';
		assertFails: ' ( 1 2  )'; "Should not read prepended spaces"
		assertFails: '( 2. 1) ';
		assertFails: '( {a} 1 )  ';
		assertFails: '( { 1 } 1 )  '.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 12/1/2010 13:44'!
testBinaryExpression
	
	self
		currentRule: #binaryExpression;
		
		assertMatches: 'a+b' andProducesNode:  (self node:'binaryExpression' with: {  'a' . self node: 'binarySelector' with: { #+ } . 'b'} );
		assertMatches: '1 * b' andProducesNode: (self node:'binaryExpression' with: {  1 . self node: 'binarySelector' with: { #* } . 'b'} );
		assertMatches: '1 "Comment "* b' andProducesNode: (self node:'binaryExpression' with: 
			{ self node: 'comment' with:{ 'Comment ' } . 1 . self node: 'binarySelector' with: { #* }. 'b'});
		assertMatches: '1 \ b / c' andProducesNode: 
			(self node: 'binaryExpression' with: 
				{ self node: 'binaryExpression' with: {  1 . self node: 'binarySelector' with: { #\ } . 'b'} . self node: 'binarySelector' with: { #/ } . 'c'} );
		
		assertFails: 'a 2';
		assertFails: 'a: 2';
		assertFails: '+ 2'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 20:02'!
testBinarySelector
	"Would be cool to find all the binary messages in the image and test them..."
	
	self
		currentRule: #binarySelector;
		
		"special cases: single pipe and single minus"
		assertMatches: '|' andProduces: #|;
		assertMatches: '-' andProduces: #-;
		
		"names starting with a minus"
		assertMatches: '-*' andProduces: #-*;
		assertMatches: '-***' andProduces: #-***;
		assertMatches: '-*%&' andProduces: #-*%&;
		assertMatches: '-*/\' andProduces: #-*/\;
		assertMatches: '-!!?' andProduces: #-!!?;
		assertMatches: '-==' andProduces: #-==;
		
		"mixed characters, check each one"
		assertMatches: '*+' andProduces: #*+;
		assertMatches: '/=\' andProduces: #/=\;
		assertMatches: '~,,,<' andProduces: #~,,,<;
		assertMatches: '>@' andProduces: #>@;
		assertMatches: '%&' andProduces: #%&;
		assertMatches: '?!!' andProduces: #?!!;
		assertMatches: '``' andProduces: #``;
		assertMatches: '+??=' andProduces: #+??=;
		assertMatches: ',!!!!,' andProduces: #,!!!!,;
		
		"Not allowed things"
		assertFails: '';
		assertFails: ' ';
		assertFails: 'w';
		assertFails: '2';
		assertFails: '--';
		assertFails: '||';
		assertFails: '|--';
		assertFails: '%--';
		assertFails: '%||';
		assertFails: '&-&-';
		assertFails: '- %';
		assertFails: ' -';
		assertFails: '& &';
		assertFails: 'w a';
		assertFails: '- -';
		assertFails: '#a a';
		assertFails: '#aa 2'.! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'ag 12/1/2010 13:37'!
testBlock
	
	self
		currentRule: #block;
		
		assertMatches: '[]' andProducesNode: { };
		assertMatches: '[  ]' andProducesNode: { };
		assertMatches: '[||]' andProducesNode: { self node: 'temporaries' };
		assertMatches: '[:a|||]' andProducesNode: { self node: 'arguments' with: { 'a' }. self node: 'temporaries' };
		assertMatches: '[||a]' andProducesNode: { self node: 'temporaries'. self node: 'statements' with: { self node: 'expression' with: { 'a' } } };
		assertMatches: '[:a|||a]' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'temporaries'. self node: 'statements' with: { self node: 'expression' with: { 'a' } } };
		assertMatches: '[:a||a|a]' andProducesNode: 
			{ self node: 'arguments' with: { 'a' }. self node: 'temporaries' with: { 'a' }. self node: 'statements' with: { self node: 'expression' with: { 'a' } } };
		assertMatches: '[:a|a|a|a]' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'statements' with:
				{ self node: 'expression' with: { self node: 'binaryExpression' with: { self node: 'binaryExpression' with: 
					{ 'a'. self node: 'binarySelector' with: { #| }. 'a' }. self node: 'binarySelector' with: { #| }. 'a' }}}};
		assertMatches: '[ :a ]  ' andProducesNode: { self node: 'arguments' with: { 'a' } };
		assertMatches: '   [ :a ]  ' andProducesNode: { self node: 'arguments' with: { 'a' } };
		assertMatches: '[ :a :b ]  ' andProducesNode: { self node: 'arguments' with: { 'a'. 'b' } };
		assertMatches: '[ :a     :b | ]' andProducesNode: { self node: 'arguments' with: { 'a'. 'b' }  };
		assertMatches: '  [   :a |   |  a  |   ]  ' andProducesNode: { self node: 'arguments' with: { 'a' }. self node: 'temporaries' with: { 'a' } };
		assertMatches: '[  :a  | a ]  ' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'statements' with: { self node: 'expression' with: { 'a' } } };
		assertMatches: '[ :a  | a. ^ 1 ]' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'statements' with:
				{ self node: 'expression' with: { 'a' }. self node: 'returnExpression' with: { 1 }}};
		assertMatches: '[  :a  |  1  ]' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'statements' with: { self node: 'expression' with: { 1 } } };
		assertMatches: '[  :b  |   []  ]' andProducesNode:
			{ self node: 'arguments' with: { 'b' }. self node: 'statements' with: { self node: 'expression' with: { self node: 'block' } } };
		
		assertMatches: '"aaaaa" "aaaa"[]'  andProducesNode: { self node:'comment' with: {'aaaaaaaaa' } };
		assertMatches: '"aaaaa" [ "a aaa"]'  andProducesNode: { self node:'comment' with: {'aaaaa' } . self node:'comment' with: {'a aaa' }};
		assertMatches: '[ "a aaa"] '  andProducesNode: { self node:'comment' with: {'a aaa' } };
		assertMatches: '[ "aaaa"] '  andProducesNode: { self node:'comment' with: {'aaaa' } };

		assertFails: '[]]';
		assertFails: '[[]';
		assertFails: '[::a]';
		assertFails: '[ :a | b';
		assertFails: '[ :a |||| b]';
		assertFails: '[ | :a | b]';
		assertFails: '[ || :a | b]';
		assertFails: '[ :a || b]';
		assertFails: '[ :a | :b  ]';
		assertFails: '[ :a || :b|]';
		assertFails: '[ :a | b| :a]';
		assertFails: '[ :a :1 | b]';
		assertFails: '[ aaaa"] ';
		assertFails: '[ "aaa"a"] ';
		assertFails: '[ "aaaa] '.
! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'SW 11/30/2010 15:53'!
testBraceExpression
	
	self
		currentRule: #braceExpression;
		
		assertMatches: '{}' andProducesNode: { };
		assertMatches: '  {  }   ' andProducesNode: { };
		assertMatches: '{1}' andProducesNode: { self node: 'expression' with: { 1 } };
		assertMatches: '  {  11 }  ' andProducesNode: { self node: 'expression' with: { 11 } };
		assertMatches: '  {  11. }  ' andProducesNode: { self node: 'expression' with: { 11 } };
		assertMatches: '{ 1.   2}' andProducesNode: { self node: 'expression' with: { 1 }. self node: 'expression' with: { 2 } };
		assertMatches: '{1.2}' andProducesNode: { self node: 'expression' with: { 1.2 } };
		assertMatches: '{1. 2}' andProducesNode: { self node: 'expression' with: { 1 }. self node: 'expression' with: { 2 } };
		assertMatches: '{1. ''.'' .}' andProducesNode: { self node: 'expression' with: { 1 }. self node: 'expression' with: { self node: 'string' with: { '.'} } };
		assertMatches: '{1  .  2  .}' andProducesNode: { self node: 'expression' with: { 1 }. self node: 'expression' with: { 2 } };
		
		assertMatches: '"aaa"{}' andProducesNode: { self node: 'comment' with: {'aaa' } };
		assertMatches: '"aaa"{ "bbb"}' andProducesNode: { self node: 'comment' with: {'aaabbb' } };			
		assertMatches: '{ "bbb"}' andProducesNode: { self node: 'comment' with: {'bbb' } };								
		
		assertFails: '.{}.';
		assertFails: '{1 .. 2}';
		assertFails: '{ 1  . 2 ..}';
		assertFails: '{.1.2}';
		assertFails: '{..1.}';
 		assertFails: '{ "bb}';
 		assertFails: '{ "b "   .  b"}'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/24/2010 11:08'!
testByteArray
	
	self
		currentRule: #byteArray;
		
		assertMatches: '#[1]' andProduces: #[1];
		assertMatches: '#  [  1  ]  ' andProduces: #[1];
		assertMatches: '#[]' andProduces: #[];
		assertMatches: '#   [1 2 3 ]' andProduces: #[1 2 3 ];
		assertMatches: '#[255 000 000 123 255]' andProduces: #[255 0 0 123 255];

		assertFails: '';
		assertFails: ' ';
		assertFails: '#[ 1';
		assertFails: '  #[ 1 ]';
		assertFails: '# [ 1.2 ]';
		assertFails: '#[ 256 1 00]';
		assertFails: '#[ -1 ]';
		assertFails: '#[ +1 ]';
		assertFails: '#[ #[ 1 ] ]  ';
		assertFails: '#( 255 )'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'SW 12/1/2010 11:53'!
testCascadedExpression
	
	self
		currentRule: #cascadedExpression;
		assertMatches: 'Transcript  show:2;cr' andProducesNode: (self node: 'cascadedExpression' with:
				{'Transcript'.  self node: 'keywordExpression' with: {self node: 'keyword' with: {#show: . 2}} .self node:'cascade' with:{#cr} } );
		assertMatches: 'Transcript  show:2 "Comment";cr' andProducesNode: (self node: 'cascadedExpression' with:
				{'Transcript'.  self node: 'keywordExpression' with: {self node: 'keyword' with: {#show: . 2}} .self node:'cascade' with:{self node:'comment' with:{'Comment'} . #cr} } );

		assertFails: 'Transcript  show:2 "Comment;cr';				
		assertFails: 'Transcript show:1'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/23/2010 18:20'!
testCharacter
	
	self
		currentRule: #character;
		
		assertMatches: '$c' andProduces: $c;
		assertMatches: '$&' andProduces: $&;
		assertMatches: '$ ' andProduces: $ ;
		assertMatches: '$`' andProduces: $`;
		assertMatches: 'Character space' andProduces: $ ;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: 'a';
		assertFails: '`';
		assertFails: ' $a';
		assertFails: '$$$';
		assertFails: '$  a';
		assertFails: '$av';
		assertFails: '$ a'.! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'ag 12/1/2010 12:12'!
testExpression
	
	self
		currentRule: #expression;
		
		assertMatches: 'a := 1';
		assertMatches: 'a := b := 1';
		assertMatches: 'a := b := c := d';
		assertMatches: '   a     :=      a     ';
		assertMatches: '   a     _   1 ';
		assertMatches: 'a:=1';
		assertMatches: 'a:=1.0';
		assertMatches: 'a:=b:=1.0';
		assertMatches: 'a_v_1';
		
		assertFails: 'a := 1 := b';
		assertFails: 'a: := b';
		assertFails: 'a: _ b';
		assertFails: 'a :_ b';
		assertFails: 'a := 1 := 2:=f';
		assertFails: 'a __ b';
		assertFails: 'a : = b'.! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'ag 11/21/2010 20:03'!
testExpressionAssignment
	
	self
		currentRule: #expression;
		
		assertMatches: 'a' andProducesNode: { 'a' };
		assertMatches: 'a := b' andProducesNode: { self node: 'variableAssignments' with: { 'a' }. 'b' };
		assertMatches: '  a   :=   b  ' andProducesNode: { self node: 'variableAssignments' with: { 'a' }. 'b' };
		assertMatches: 'a := b := c := 1' andProducesNode: { self node: 'variableAssignments' with: { 'a'. 'b'. 'c' }. 1 };
		assertMatches: '  a    :=  b  :=  c :=   1' andProducesNode: { self node: 'variableAssignments' with: { 'a'. 'b'. 'c' }. 1 };
		assertMatches: 'a _ b' andProducesNode: { self node: 'variableAssignments' with: { 'a' }. 'b' };
		
		assertFails: 'a : = b';
		assertFails: ' := b';
		assertFails: 'a := ';
		assertFails: 'a := b :=';
		assertFails: 'a := b := 1 := a';
		assertFails: 'a := := b   ';
		assertFails: 'a _ _ b';
		assertFails: 'a __ b';
		assertFails: 'a :=:= b';
		assertFails: 'a := b v := a';
		assertFails: 'a  b := a '.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 12/1/2010 12:02'!
testIdentifier
	
	self
		currentRule: #identifier;
		
		assertMatches: 'abc' andProduces: 'abc';
		assertMatches: 'a1' andProduces: 'a1';
		assertMatches: '1' andProduces: '1';
		assertMatches: 'aaaaaaaa1233333' andProduces: 'aaaaaaaa1233333';
		assertMatches: '' andProduces: '';
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '|';
		assertFails: ':';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc:';
		assertFails: 'abc<<'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/22/2010 11:30'!
testKeywordSelector
	
	self
		currentRule: #keyword;
		
		assertMatches: 'abc:' andProduces: #abc:;
		assertMatches: 'a1:' andProduces: #a1:;
		assertMatches: '1:' andProduces: #1:;
		assertMatches: 'aaaaaaaa1233333:' andProduces: #aaaaaaaa1233333:;
		assertMatches: ':' andProduces: #:;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc';
		assertFails: 'abc<<'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 12/1/2010 11:16'!
testNumber
	
	"TODO extend test, when numberS2J gets extended (e.g. -2 0.2 etc.)"
	
	self
		currentRule: #numberS2J;
		
		assertMatches: '1' andProduces: 1;
		assertMatches: '0' andProduces: 0;
		assertMatches: '123' andProduces: 123;
		assertMatches: '112345678' andProduces: 112345678;
		assertMatches: '000' andProduces: 0;
		assertMatches: '-000' andProduces: 0;
		assertMatches: '-1' andProduces: -1;
		assertMatches: '-1.22' andProduces: -1.22;
		assertMatches: '1.22e23' andProduces: 1.22e23;
		assertMatches: '122e23' andProduces: 12200000000000000000000000;
		assertMatches: '35r35' andProduces: 110;
		assertMatches: '-35r35' andProduces: -110;
		assertMatches: '22r12.12' andProduces: 22r12.12;
		assertMatches: '22r12.12e12' andProduces: 22r12.12e12;
		assertMatches: '-22r-12.12e12' andProduces: -22r-12.12e12;
		assertMatches: '-22r-12.12e12' andProduces: -22r-12.12e12;
		assertMatches: '22r-12.12e12' andProduces: 22r-12.12e12;
		assertMatches: '-22r12.12e12' andProduces: -22r12.12e12;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '+1';
		assertFails: '+1.2';
		assertFails: '+23r23.3';
		assertFails: '+23r+233.3';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: 'abc';
		assertFails: '<<'.! !

!S2JParserTest methodsFor: 'test-random' stamp: 'ag 11/30/2010 16:02'!
testRandomMethods
	"Match some random methods from the image.
	>>findUnmatchableMethod should match all methods in this image before running this test.
	This test should always succeed, failures are printed to the Transcript, because testing random methods
	is not a real Unit-test."
	
	| result |
	result := self class matchRandomMethods: 10.
	Transcript cr; cr; show: result.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/24/2010 10:47'!
testString
	
	"TODO extend test, if stringS2J gets extended to support '' inside of a string"
	
	self
		currentRule: #stringS2J;
		
		assertMatches: '''a''' andProduces: (self node: 'string' with: { 'a' });
		assertMatches: '''''' andProduces: (self node: 'string' with: { '' });
		assertMatches: '''123''' andProduces: (self node: 'string' with: { '123' });
		assertMatches: ''' '''' '''' ''' andProduces: (self node: 'string' with: { ' '' '' ' });
		assertMatches: '''''''123''' andProduces: (self node: 'string' with: { '''123' });
		assertMatches: '''    ''' andProduces: (self node: 'string' with: { '    ' });
		assertMatches: '''"$%&''' andProduces: (self node: 'string' with: { '"$%&' });
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '#abc';
		assertFails: ''' ''a';
		assertFails: ' ''''';
		assertFails: 'a''''';
		assertFails: '''as';
		assertFails: 'as''';
		assertFails: ''' '' ''';
		assertFails: '''a '' a''';
		assertFails: '''/'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/23/2010 16:34'!
testSymbol
	
	"TODO extend test, if stringS2J gets extended to support '' inside of a string"
	
	self
		currentRule: #symbolS2J;
		
		assertMatches: '#abc' andProduces: #abc;
		assertMatches: '#abc1' andProduces: #abc1;
		assertMatches: '#a' andProduces: #a;
		assertMatches: '#''a''' andProduces: #a;
		assertMatches: '#''a b c''' andProduces: #'a b c';
		assertMatches: '#''''' andProduces: #'';
		assertMatches: '#''123''' andProduces: #'123';
		assertMatches: '#''    ''' andProduces: #'    ';
		assertMatches: '#''"$%&''' andProduces: #'"$%&';
		assertMatches: '#abc1:' andProduces: #abc1:;
		assertMatches: '#|' andProduces: #|;
		assertMatches: '#-&&' andProduces: #-&&;
		assertMatches: '#-' andProduces: #-;
		assertMatches: '#a:b:c1' andProduces: #a:b:c1;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: 'abc';
		assertFails: '#1abc';
		assertFails: '#""';
		assertFails: '#'' ''a';
		assertFails: '# ''''';
		assertFails: '#a''''';
		assertFails: '#''as';
		assertFails: '#as''';
		assertFails: '#'' '' ''';
		assertFails: '#''a '' a''';
		assertFails: '#123';
		assertFails: '###';
		assertFails: '#''/'.! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'ag 11/24/2010 10:47'!
testUnaryExpression
	
	self
		currentRule: #unaryExpression;
		
		assertMatches: 'a b' andProducesNode: { 'a'. #b };
		assertMatches: '1 b' andProducesNode: { 1. #b };
		assertMatches: 'a b c' andProducesNode: { self node: 'unaryExpression' with: { 'a'. #b }. #c };
		assertMatches: ' a  b c d ' andProducesNode: { self node: 'unaryExpression' with: { self node: 'unaryExpression' with: { 'a'. #b }. #c }. #d };
		assertMatches: '   ''abc''    b1     ' andProducesNode: { self node: 'string' with: { 'abc' }. #b1 };
		
		assertFails: '';
		assertFails: '  ';
		assertFails: 'a';
		assertFails: '  a  ';
		assertFails: 'a  1';
		assertFails: 'a b 1';
		assertFails: 'a || b';
		assertFails: 'a % b';
		assertFails: 'a -|';
		assertFails: 'a b > a';
		assertFails: 'a unary>'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 16:06'!
testUnarySelector
	"Would be cool to find all the unary messages in the image and test them..."
	
	self
		currentRule: #unarySelector;
		
		assertMatches: 'abc' andProduces: #abc;
		assertMatches: 'a1' andProduces: #a1;
		assertMatches: '1' andProduces: #1;
		assertMatches: 'aaaaaaaa1233333' andProduces: #aaaaaaaa1233333;
		assertMatches: '' andProduces: #;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '|';
		assertFails: ':';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc:';
		assertFails: 'abc<<'.! !

TestCase subclass: #S2JTranslatorTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JTranslatorTest methodsFor: 'helpers' stamp: 'HK 11/29/2010 16:21'!
assertTranslationOf: aString equals: anotherString

	self
		assert: (S2JTranslator translateString: aString)
		equals: anotherString.! !

!S2JTranslatorTest methodsFor: 'helpers' stamp: 'HK 11/29/2010 16:28'!
expectedFailures

	^ Array
		with: #testAddition
		with: #testLoop
		with: #testTrueFalse
		with: #testNonLocalReturn
		with: #testClientEvaluation.! !

!S2JTranslatorTest methodsFor: 'testing' stamp: 'HK 11/29/2010 16:19'!
testAddition

	self
		assertTranslationOf: '3+4'
		equals: '3+4'.

! !

!S2JTranslatorTest methodsFor: 'test stories' stamp: 'HK 11/29/2010 16:31'!
testClientEvaluation

	self fail.
	"clientEval: [ JsGlobal alert: 3+4]"! !

!S2JTranslatorTest methodsFor: 'test stories' stamp: 'HK 11/29/2010 16:21'!
testLoop

	self
		assertTranslationOf: '{1. 2. 3} do: [:i | JSGlobal alert: i]'
		equals:
			'var arr = [1, 2, 3]; for (var i=0; i<arr.length; i++) { alert(i);}'
			
			"[1, 2, 3].do(function(i) { alert(i) })"

! !

!S2JTranslatorTest methodsFor: 'test stories' stamp: 'HK 11/29/2010 16:22'!
testNonLocalReturn
	
	self
		assertTranslationOf: '
			[ JsGlobal alert: ''outer''.
			[JsGlobal alert: ''inner''.
			 ^self].
			JsGlobal alert: ''end'' ]'
		equals: ''
		"to be continued"! !

!S2JTranslatorTest methodsFor: 'test stories' stamp: 'LT 11/30/2010 12:26:45.243'!
testSelectorTranslation

	| squeakSelector jsFunctionname |
	squeakSelector := 'do:forDuration:'.
	jsFunctionname := (S2JTranslator squeakSelectorToJSFunctionname: squeakSelector).
	self assert: 'do_forDuration_' equals: jsFunctionname.
	squeakSelector := '++'.
	jsFunctionname := (S2JTranslator squeakSelectorToJSFunctionname: squeakSelector).
	self assert: 'PLUSPLUS' equals: jsFunctionname.! !

!S2JTranslatorTest methodsFor: 'test stories' stamp: 'HK 11/29/2010 16:21'!
testTrueFalse

	self
		assertTranslationOf: '10 < 5 ifTrue: [ JsGlobal alert: ''small'' ] ifFalse: [ JsGlobal alert: ''big'' ]'
		equals:
			'if (10 < 5) { alert(''small''} else { alert(''big'') }'! !

OMeta2 subclass: #S2JParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParser commentStamp: 'ag 11/30/2010 16:20' prior: 0!
tokens
	Rules, that read one single element in squeal source-code. They all don't return an S2JNode, all are parsed to the
	according Squeak-object (e.g. Number, Symbol, String).
	Spaces before and after the tokens in the source-code are NOT read!!
	This is because in some cases, tokens must be read exactly, without emitting any spaces.
	The things returned from here must not be converted anymore, especially the <> operator is unnecessary.
	
complex elements
	Bigger parts of squeak source-code. Most return an S2JNode (using #node:with:), which contains further complex elements or object read
	by rules in the tokens-protocol. As convention, the nodes (except for helper-rules like primary and messagePattern) should be named like the rules.
	Each rule is supposed to read spaces before applying itself!! This way, rules from this protocol can be called serially, without
	reading spaces inbetween. However, spaces must be read between single tokens.
	
helpers
	Squeak-methods, that are called from within ometa-rules (or helper-rules).
	OMeta2 return-parts of the rules (righthand of ->) cannot contain multiple expressions (?)...


__ Issues __ (Things, that cannot be parsed; there is no Specification for the Squeak-syntax, so new thigs can appear at any time)


!

!S2JParser class methodsFor: 'matching' stamp: 'ag 11/30/2010 16:09'!
matchMethod: aCompiledMethod
	
	^ self
		matchString: aCompiledMethod getSource asString
		with: #method! !

!S2JParser class methodsFor: 'matching' stamp: 'LT 11/29/2010 12:26:51.13'!
matchMethod: aStringOrSymbol of: aClassOrClassname
	
	| method |
	method := self method: aStringOrSymbol of: aClassOrClassname.
	^ method
		ifNotNil: [ self matchMethod: method ]
		ifNil: [ self error: 'Method ', aStringOrSymbol asString, ' not found in class ', aClassOrClassname asString ]! !

!S2JParser class methodsFor: 'matching' stamp: 'ag 11/23/2010 15:51'!
matchString: aString with: aRule
	
	| matcherToResult |
	matcherToResult := self matchStringWithoutEnd: aString with: aRule.
	matcherToResult key apply: #end.
	^ matcherToResult value! !

!S2JParser class methodsFor: 'matching private' stamp: 'ag 11/23/2010 15:49'!
matchStringExplore: aString with: aRule
	
	| matcherToResult |
	matcherToResult := self matchStringWithoutEnd: aString with: aRule.
	[ matcherToResult key apply: #end ]
		on: OM2Fail do: [ :err | matcherToResult value explore. self error: err ].
	^ matcherToResult value! !

!S2JParser class methodsFor: 'matching private' stamp: 'LT 11/29/2010 12:50:54.628'!
matchStringWithoutEnd: aString with: aRule
	"Returns both the matcher and the result as an association."
	
	| matcher result |
	matcher := self matcherOn: aString readStream.
	result := matcher
		apply: #empty;
		apply: aRule.
	matcher apply: #spaces.
	^ matcher -> result! !

!S2JParser class methodsFor: 'lookup' stamp: 'LT 11/30/2010 12:15:11.967'!
method: aStringOrSymbol of: aClassOrClassname
	
	^ (Smalltalk classNamed: aClassOrClassname asString)
		lookupSelector: aStringOrSymbol asSymbol! !

!S2JParser class methodsFor: 'parsing' stamp: 'ag 12/1/2010 11:04'!
parseNumber: aString
	"Parse aString to an instance of Number. Return nil, if parsing failed.
	Return nil also, if a number could be read, but not the whole String was processed.
	(Number readFrom: aString does not fail in that case)."
	
	| stream result |
	stream := ReadStream on: aString.
	result := Number readFrom: stream ifFail: [ nil ].
	stream atEnd ifFalse: [ result := nil ].
	^ result! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 12/1/2010 10:18'!
array =
	
	$(
		(
			(spaces
				(((($# spaces)? array:array) -> [ array ])
				| (symbolBase:symbol -> [ symbol asSymbol ])
				| literalToken
				):element
			) -> [ element ]
		)*:contents
	ch($))
	-> [ contents asArray ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 12/1/2010 13:35'!
binaryExpression =
	
	(binaryExpression | unaryExpression | primary):receiver
	spaces:comment binarySelector:selector
	(unaryExpression | primary):argument
	
	-> [ self node: 'binaryExpression' withoutNils: 
		{ self extractCommentOf:comment. receiver. 
			self node: 'binarySelector' with: { selector }. argument } ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 12/1/2010 13:32'!
binarySelector =
	
	<$| | ((specialCharacter | $-) specialCharacter*)>:selector
	-> [ selector asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'SW 11/30/2010 15:47'!
block =
	
	ch($[):comment1
		( (ch($:) identifier:argumentName -> [ argumentName ])+:arguments ch($|)?:argumentSeparator )?
		temporaries?:temporaries
		statements?:statements
	ch($]):comment2
	
	// Forbidden (and not covered above): Arguments without separator bar, but with statements or temporaries
	?[ argumentSeparator notNil | arguments isNil | ((temporaries isNil or: [temporaries isEmpty]) & (statements isNil or: [statements isEmpty])) ]
	
	-> [ self node: 'block' withoutNils: { self extractCommentOf:comment1 . 
			self node: 'arguments' withoutNilsNotEmpty: arguments.
			temporaries. statements .
			self extractCommentOf:comment2} ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'SW 11/30/2010 15:49'!
braceExpression =
	
	ch(${):comment1
		(expression:expr ch($.) -> [ expr ] )*:expressions
		expression?:lastExpression
	ch($}):comment2
	
	-> [ self braceExpressionWith: expressions and: lastExpression and: (self extractCommentOf:comment1,  comment2) ]! !

!S2JParser methodsFor: 'helpers' stamp: 'SW 11/30/2010 15:50'!
braceExpressionWith: expressions and: lastExpression and:comment
	
	^ self node: 'braceExpression' withoutNils: 
		((expressions
			ifNotNil: [ expressions ]
			ifNil: [ {} ])
				asOrderedCollection
					add: lastExpression;
					addFirst: comment;
					yourself)! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/23/2010 15:43'!
byteArray =
	
	$# ch($[)
		((spaces numberS2J:byte ?[ (byte isInteger) & (byte <= 255) & (byte >= 0)] ) -> [ byte ])*:contents
	ch($])
	-> [ contents asByteArray ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'SW 12/1/2010 11:23'!
cascadedExpression =
	// Constructs a node, whose first node is the receiver and all further nodes are messages sent to that receiver
	
	(keywordExpression | binaryExpression | unaryExpression):mainExpression
	(
		ch($;):comment1
		(
		((spaces binarySelector:selector (unaryExpression | primary):argument) -> [ self node: 'binary' with: { selector. argument }])
		| (((spaces keyword:selector (binaryExpression | unaryExpression | primary):argument) -> 
				[ selector -> argument])+:selectorsAndArguments -> [ self keywordMessage: selectorsAndArguments ])
		| (spaces unarySelector:selector -> [ self node: 'unary' with: { selector } ])
		):cascade
		-> [self node:'cascade' withoutNils:comment1,cascade ]
	)+:cascades
	-> [ self cascadedExpression: mainExpression with: cascades ]! !

!S2JParser methodsFor: 'helpers' stamp: 'SW 11/30/2010 16:46'!
cascadedExpression: mainExpression with: cascades
	"The receiver of the mainExpression is the receiver of the whole cascade.
	It is the first node, the other nodes are the messages sent to that receiver"
	
	^ self node: 'cascadedExpression' withoutNils: 
		((cascades
			ifNotNil: [ cascades ]
			ifNil: [ {} ])
				asOrderedCollection
					addFirst: mainExpression;
					addFirst: mainExpression removeFirst;
					yourself)! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/30/2010 11:01'!
ch :c =
	// This reads a given element and all spaces before it and return the element
	// ch was supposed to mean character, but this can be applied to any object
	// Called like this: ch($=)
	
	spaces:comment exactly(c)
	-> [ comment ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/24/2010 17:22'!
character =
	// The method in the super class is #char, don't confuse it!!
	
	( ($$ <anything>:value) | (seq('Character space') -> [ $  ]):value )
	-> [ value asCharacter ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'LT 12/1/2010 11:48:21.177'!
expression =
	
	((spaces identifier:identifier (token(':=') | ch($_))) -> [ identifier ])*:variableNames 
	(cascadedExpression | keywordExpression | binaryExpression | unaryExpression | primary):expressionBody
	
	// Create an extra node for the assignments, if any
	-> [ self node: 'expression' withoutNils: { 
			(self node: 'variableAssignments' withoutNilsNotEmpty: variableNames).
			expressionBody} ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 12/1/2010 13:34'!
extractCommentOf: anOrderedCollection
	
	| result |
	result := ''.
	anOrderedCollection do: [:item |
		item name = 'comment' ifTrue: [
			result := result , item first
			]].
	^ result isEmpty
		ifTrue: [nil]
		ifFalse: [self node: 'comment' with:{result}]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 12/1/2010 11:57'!
identifier =
	
	identifierBase:result
	-> [ result ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/30/2010 11:03'!
identifierBase =
	
	<letter letterOrDigit*>! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/22/2010 16:34'!
keyword =
	
	<identifierBase $:>:selector
	-> [ selector asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/24/2010 17:28'!
keywordExpression =
	
	(binaryExpression | unaryExpression | primary):receiver
	((((spaces keyword:selector ( binaryExpression | unaryExpression | primary ):argument) -> [ selector -> argument ])+:selectorAndArguments)
		-> [ self keywordMessage: selectorAndArguments ]):message
	-> [self node: 'keywordExpression' with: { receiver. message}]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/22/2010 11:20'!
keywordMessage: selectorAndArgumentNames
	"selectorAndArgumentNames contains associations, the part of the selector and the argument name.
	Resulting S2JNode will have the selector as symbol on the first position and the argumentNames at the other positions."
	
	| selector result |
	selector := ''.
	result := self node: 'keyword' with: {}.
	selectorAndArgumentNames do: [ :each |
		selector := selector, each key.
		result add: each value].
	^ result
		addFirst: selector asSymbol;
		yourself! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/30/2010 19:08'!
literalToken =
	// Things, that can be evalutated to an object without any context (array-literals excluded)
	
	numberS2J
	| symbolS2J
	| character
	| stringS2J
	| byteArray! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/30/2010 10:51'!
messagePattern =
	
	((spaces keyword:selectorPart spaces identifier:argumentName) -> [ selectorPart -> argumentName ])+:selectorAndArgumentNames
	-> [ self keywordMessage: selectorAndArgumentNames ]
|	spaces binarySelector:selector spaces identifier:argumentName
	-> [ self node: 'binary' with: { selector. argumentName } ]	
|	spaces identifier:unarySelector
	-> [ self node: 'unary' with: { unarySelector } ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'LT 11/29/2010 13:36:37.039'!
method =
	
	messagePattern:head
	temporaries?:temps
	primitiveDeclaration?:primitive
	statements?:statements
	
	-> [ self node: 'method' withoutNils: { head. temps. primitive. statements } ]! !

!S2JParser methodsFor: 'helpers' stamp: 'LT 11/29/2010 13:34:35.706'!
node: aString with: anOrderedCollection
	
	^ S2JNode named: aString with: anOrderedCollection! !

!S2JParser methodsFor: 'helpers' stamp: 'LT 11/29/2010 13:29:39.329'!
node: aString withoutNils: anOrderedCollection
	
	^ S2JNode named: aString withoutNils: anOrderedCollection! !

!S2JParser methodsFor: 'helpers' stamp: 'LT 11/29/2010 13:23:15.597'!
node: aString withoutNilsNotEmpty: anOrderedCollection
	
	^ S2JNode named: aString withoutNilsNotEmpty: anOrderedCollection! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 12/1/2010 10:54'!
numberS2J =
	// Called that way because of super-class-method
	// The string is parsed once, assuring, that
	
	<
		($-? digit+ $r)?
		$-?
		(digit | upper)+
		($. (digit | upper)+)?
		($e ($- | $+)? (digit | upper)+)?
	>:result
	?[ (self class parseNumber: result) notNil ]
	-> [ self class parseNumber: result ]
	
	// An alternative way to evaluate a number...
	| ($# <digit>:value) -> [ value asNumber ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/30/2010 19:08'!
primary =
	// something, that evaluates to an object
	
	spaces (literalToken | reservedWord | (($# spaces array:array -> [ array ])) | identifier):result -> [ result ]
|	block
|	braceExpression
|	token('(') expression:expr token(')') -> [ expr name: 'paranthesizedExpression'; yourself ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/24/2010 17:30'!
primitiveDeclaration =
	
	ch($<)
		token('primitive:')
		spaces 
		(
			(numberS2J:primitiveNumber ?[ primitiveNumber isInteger ])
		|	(stringS2J:methodName (token('module:') spaces stringS2J:moduleName))?
		)
	ch($>)
	-> [ self node: 'primitiveDeclaration' withoutNils: { primitiveNumber. methodName. moduleName } ]! !

!S2JParser methodsFor: 'helpers' stamp: 'SW 11/30/2010 11:30'!
removeNils: anOrderedCollection
	
	| result |
	result := anOrderedCollection reject: [ :each | each isNil].
	^ result isEmpty
		ifTrue: [ nil ]
		ifFalse: [ result ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/30/2010 19:06'!
reservedWord =
	
	<(seq('true') | seq('false') | seq('nil') | seq('self') | seq('super') | seq('thisContext') | seq('homeContext'))>:reservedWord
	-> [ self node: 'reservedWord' with: { reservedWord }]! !

!S2JParser methodsFor: 'tokens' stamp: 'SW 11/30/2010 11:32'!
space =
	// A comment is equivalent to a space-character.
	// spaces are ignored (return nil), but comments result in a special node and are collected
	
	^space -> [ nil ]
	| (($" <(~$" anything)*>:result $") -> [ self node: 'comment' with: { result } ])! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
specialCharacter =
	
	$+ | $* | $/ | $\ | $~ | $< | $> | $= | $@ | $% | $& | $? | $!! | $` | $,! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/24/2010 17:31'!
statements =
	
	(expression:expr ch($.) -> [ expr ] )*:expressions
	(ch($^)?:returnSign expression:lastExpression ch($.)? ~expression)?
	
	-> [ self statementsWithExpressions: expressions lastExpression: lastExpression returnSign: returnSign ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/21/2010 20:35'!
statementsWithExpressions: expressions lastExpression: lastExpression returnSign: returnSign
	"Build the contents of a statements-node from the information gathered while parsing 'statements'"
	
	"Rename the last expression, if it was a return-expression"
	lastExpression notNil & returnSign notNil ifTrue: [ lastExpression name: 'returnExpression' ].
	expressions add: lastExpression.
	^ self node: 'statements' withoutNilsNotEmpty: expressions! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/24/2010 10:40'!
stringS2J =
	// Called that way because of super-class-method
	
	$' ((~$' anything) | (($' $') -> [ $' ]))*:values $'
	-> [ self node: 'string' with: { values asOrderedCollection inject: '' into: [ :a :b | a := a, b asString ] } ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 12/1/2010 13:33'!
symbolBase =
	// This does not return a symbol, just consume input. Inside of array, symbols can be represented without
	// the #-character
	
	binarySelector
|	<keyword+ identifier>
|	<keyword+>
|	identifier! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 12/1/2010 10:15'!
symbolS2J =
	// Called that way because of super-class-method
	
	$# (symbolBase | (stringS2J:stringSymbol -> [ stringSymbol first ])):value
	-> [ value asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:10'!
temporaries =
	
	ch($|)
		(spaces identifier:identifier -> [ identifier ])*:variableNames
	ch($|)
	-> [ self node: 'temporaries' with: variableNames ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/22/2010 11:11'!
unaryExpression =
	
	(unaryExpression | primary):receiver spaces unarySelector:message
	-> [ self node: 'unaryExpression' with: {receiver. message} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 12/1/2010 11:57'!
unarySelector =
	
	identifier:selector ~$:
	-> [ selector asSymbol ]! !

OMeta2 subclass: #S2JTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

S2JTranslator subclass: #S2JLibaryTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

S2JTranslator subclass: #S2JSyntaxTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JSyntaxTranslator methodsFor: 'as yet unclassified' stamp: 'sec 11/30/2010 14:41'!
plus =

	{ anythingS2J:receiver symbol:selector anythingS2J:argument }:node isNode(node, 'binaryExpression')
	
	// -> [ (self binarySyntax: receiver selector: selector argument: argument) ifNil: [ self fail ]]
! !

!S2JTranslator class methodsFor: 'mapping' stamp: 'LT 11/30/2010 12:06:42.591'!
indentifierForSqueakCharacter: aCharacter 

	^ self squeakCharacterToJSIdentifier at: aCharacter ifAbsent: [nil]! !

!S2JTranslator class methodsFor: 'mapping' stamp: 'LT 12/1/2010 14:36:40.378'!
squeakCharacterToJSIdentifier
	"Several characters can be used for selectors in squeak but
	are not allowed for JavaScript functionname"
	
	^ Dictionary newFrom:
		{$: -> '$'. 
		$+ -> '$plus'.
		$- -> '$minus'.
		$* -> '$times'.
		$/ ->  '$slash'.
		$\ -> '$backslash'.
		$~ -> '$tilde'.
		$< -> '$less'.
		$> -> '$greater'.
		$= -> '$equals'.
		$@ -> '$at'.
		$% -> '$percent'.
		$& -> '$and'.
		$? -> '$question'.
		$!! -> '$exclamation'.
		$` -> '$backtick'.
		$, -> '$comma'. 
		$| -> '$pipe'}! !

!S2JTranslator class methodsFor: 'mapping' stamp: 'LT 12/1/2010 14:40:07.705'!
squeakSelectorToJSFunctionName: aStringOrSymbol	
	
	| result |
	result := ''.
	aStringOrSymbol asString do: [ :each || jsIdentifier |
		jsIdentifier := self indentifierForSqueakCharacter: each.
		jsIdentifier
			ifNotNil: [result := result, jsIdentifier ]
			ifNil: [result := result, each asString]].
	^ result! !

!S2JTranslator class methodsFor: 'mapping' stamp: 'LT 11/30/2010 12:08:16.095'!
squeakSelectorToJSFunctionname: aStringOrSymbol	
	
	| result |
	result := ''.
	aStringOrSymbol asString do: [ :each || jsIdentifier |
		jsIdentifier := self indentifierForSqueakCharacter: each.
		jsIdentifier
			ifNotNil: [result := result, jsIdentifier ]
			ifNil: [result := result, each asString]].
	^ result! !

!S2JTranslator class methodsFor: 'translating' stamp: 'ag 11/24/2010 18:07'!
translateMethod: aCompiledMethod
	
	^ self match: (S2JParser matchMethod: aCompiledMethod) with: #methodDeclaration! !

!S2JTranslator class methodsFor: 'translating' stamp: 'ag 11/24/2010 18:09'!
translateMethod: aStringOrSymbol of: aClassOrClassname
	
	^ self translateMethod: (S2JParser method: aStringOrSymbol of: aClassOrClassname)! !

!S2JTranslator class methodsFor: 'translating' stamp: 'SW 11/30/2010 10:43'!
translateString: aString 
	| sourceRule targetRule |
	
	sourceRule := #method.

	targetRule := #stringS2J.	
	^self translateString: aString from: sourceRule to: targetRule
	
	! !

!S2JTranslator class methodsFor: 'translating' stamp: 'ag 11/24/2010 18:06'!
translateString: aString from: sourceRule to: targetRule
	"Translate Squeak-code to JS. Parse the code using the sourceRule and translate the AST using the targetRule."
	
	^ self match: (S2JParser matchString: aString with: sourceRule) with: targetRule
	! !

!S2JTranslator methodsFor: 'as yet unclassified' stamp: 'LT 12/1/2010 11:36:57.835'!
anyExpression = 

|	trString
|	trNumber
|	trCharacter
|	trSymbol
| 	trArray
|	trBraceExpression
|	trByteArray
|	trIdentifier
|	trReservedWord
|	trUnaryExpression
|	trBinaryExpression
|	trKeywordExpression
|	trCascadedExpression
|	trBlock! !

!S2JTranslator methodsFor: 'helpers' stamp: 'sec 11/30/2010 14:37'!
isNode :anObject :aString =
	
	?[ anObject class = S2JNode and: [ anObject name = aString ]]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 11:30:44.643'!
node :aString =
	
	// This does not create an S2JNode, but reads one from the input stream and returns it!!
	// For this purpose, this is an ometa-rule with one argument, call like this: node('binarySelector')
	// Other things, that can be found in the result of S2JParser are symbols, strings, numbers etc. (already implemented in superclass)
	// Collections, that are not S2JNodes, should not occur...!!
	
	anything:node ?[node class = S2JNode and: [node name = aString]]
	-> [ node ]! !

!S2JTranslator methodsFor: 'as yet unclassified' stamp: 'LT 12/1/2010 15:54:50.349'!
trAnything = 
 
	[self halt]
	trString
|	trNumber
|	trCharacter
|	trSymbol
| 	trArray
|	trBraceExpression
|	trByteArray
|	trIdentifier
|	trReservedWord
|	trUnaryExpression
|	trBinaryExpression
|	trKeywordExpression
|	trBinarySelector
|	trCascadedExpression
|	trBlock
|	trMethod
| 	trStatements
|	trTemporaries! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 15:58:59.219'!
trArray = 
	
	{ trAnything*:translatedNode }:node ?[ node isArrayNode ]
	-> [ ' [ ', (translatedNode separatedBy: ','), ']']! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 15:53:57.973'!
trBinaryExpression =

	{ trAnything*:node } -> [ (S2JNode named: 'x' with: node) separatedBy: '' ]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:57:26.987'!
trBinarySelector = 
	
	node('binarySelector'):node -> [ self class squeakSelectorToJSFunctionName: node first ]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:16:25.361'!
trBlock =

	{ trAnything* }:node isNode(node, 'block') ! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:07:24.987'!
trBraceExpression =

	trArray! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:07:33.489'!
trByteArray =

	trArray! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:16:07.956'!
trCascadedExpression =

	{ trAnything* }:node isNode(node, 'cascadedExpression') ! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 13:38:57.658'!
trCharacter = 

	node('character'):node
	-> [ node first ]! !

!S2JTranslator methodsFor: 'as yet unclassified' stamp: 'LT 12/1/2010 14:09:11.817'!
trIdentifier =

	node('identifier'):node -> [ node first ]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:15:44.49'!
trKeywordExpression =

	{ trAnything* }:node isNode(node, 'keywordExpression') ! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:16:37.282'!
trMethod =

	{ trAnything* }:node isNode(node, 'method') ! !

!S2JTranslator methodsFor: 'translate' stamp: 'LT 12/1/2010 11:28:41.626'!
trMethodCall =
	
	node('unaryExpression')! !

!S2JTranslator methodsFor: 'translate' stamp: 'LT 12/1/2010 11:28:32.619'!
trMethodDeclaration =
	
	node('method'):method
	-> [ 'This is a JS method declaration!! Those are the contents: ', method asString ]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 15:48:31.482'!
trNumber =
	
	 // literal translation instead of number objects in JS
	number:number -> [ number asString ]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 11:27:52.395'!
trReservedWord =
	
	node('reservedWord'):node
	-> [ node first ]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:17:35.011'!
trStatements =

	{ trAnything* }:node isNode(node, 'statements') ! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 11:27:19.481'!
trString =
	// Reads one instance of String. Reads one node and returns the string in it.
	
	node('string'):node
	-> [ node first ]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 12:03:47.929'!
trSymbol =
	
	symbol:s -> [ s asString ]! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:17:49.602'!
trTemporaries =

	{ trAnything* }:node isNode(node, 'temporaries') ! !

!S2JTranslator methodsFor: 'helpers' stamp: 'LT 12/1/2010 14:15:26.642'!
trUnaryExpression =

	{ trAnything* }:node isNode(node, 'unaryExpression') ! !
